"""
Hotkeys management.
Requires "gi" package and "Gtk" & "Keybinder" modules.
"""
# pylint: disable=broad-except
import sys
import threading

try:
    # pylint: disable=import-error
    import gi
    gi.require_version('Keybinder', '3.0')  # noqa
    gi.require_version('Gtk', '3.0')  # noqa
    from gi.repository import Keybinder, Gtk
    # pylint: enable=import-error
except Exception as error:
    sys.stderr.write('Could not import Keybinder and Gtk. Error was: "{}"\n'.format(error))
    sys.stderr.write('Global shortcuts will not work.\n')
    IS_INIT = False
    ERROR = str(error)
else:
    IS_INIT = True

from clay.settings import Settings
from clay.eventhook import EventHook
from clay.notifications import NotificationArea


class HotkeyManager(object):
    """
    Manages configs.
    Runs Gtk main loop in a thread.
    """
    DEFAULT_HOTKEYS = {
        'play_pause': 'XF86AudioPlay',
        'next': 'XF86AudioNext',
        'prev': 'XF86AudioPrev'
    }

    instance = None

    def __init__(self):
        assert self.__class__.instance is None, 'Can be created only once!'
        self.hotkeys = {}
        self.config = None

        self.play_pause = EventHook()
        self.next = EventHook()
        self.prev = EventHook()

        if IS_INIT:
            Keybinder.init()
            self.initialize()

            threading.Thread(target=Gtk.main).start()
        else:
            NotificationArea.notify(
                'Could not import Keybinder and Gtk. Error was: "{}"\n'
                'Global shortcuts will not work.'.format(ERROR)
            )

    @classmethod
    def get(cls):
        """
        Create new :class:`.HotkeyManager` instance or return existing one.
        """
        if cls.instance is None:
            cls.instance = HotkeyManager()

        return cls.instance

    @staticmethod
    def load_keys():
        """
        Load hotkey config from settings.
        """
        config = Settings.get_config()
        hotkeys = config.get('hotkeys', {})
        for operation, default_key in HotkeyManager.DEFAULT_HOTKEYS.items():
            if operation not in hotkeys or not hotkeys[operation]:
                hotkeys[operation] = default_key
        return hotkeys

    def initialize(self):
        """
        Unbind previous hotkeys, re-read config & bind new hotkeys.
        """
        for operation, key in self.hotkeys.items():
            Keybinder.unbind(key)
        self.hotkeys = self.load_keys()
        for operation, key in self.hotkeys.items():
            Keybinder.bind(key, self.fire_hook, operation)

    def fire_hook(self, key, operation):
        """
        Fire hook by name.
        """
        assert key
        getattr(self, operation).fire()



"""
Components for song listing.
"""
# pylint: disable=too-many-arguments
# pylint: disable=too-many-instance-attributes
# pylint: disable=too-many-public-methods
import os
from string import digits
try:
    # Python 3.x
    from string import ascii_letters
except ImportError:
    # Python 2.3
    from string import letters as ascii_letters
import urwid
from clay.notifications import NotificationArea
from clay.player import Player
from clay.gp import GP


class SongListItem(urwid.Pile):
    """
    Widget that represents single song item.
    """
    signals = [
        'activate',
        'append-requested',
        'unappend-requested',
        'station-requested',
        'context-menu-requested'
    ]

    STATE_IDLE = 0
    STATE_LOADING = 1
    STATE_PLAYING = 2
    STATE_PAUSED = 3

    STATE_ICONS = {
        0: ' ',
        1: u'\u2505',
        2: u'\u25B6',
        3: u'\u25A0'
    }

    def __init__(self, track):
        self.track = track
        self.index = 0
        self.state = SongListItem.STATE_IDLE
        self.line1 = urwid.SelectableIcon('', cursor_position=1000)
        self.line1.set_layout('left', 'clip', None)
        self.line2 = urwid.Text('', wrap='clip')

        self.line1_wrap = urwid.AttrWrap(self.line1, 'line1')
        self.line2_wrap = urwid.AttrWrap(self.line2, 'line2')

        self.content = urwid.Pile([
            self.line1_wrap,
            self.line2_wrap,
            urwid.Text('')
        ])

        self.is_focused = False

        super(SongListItem, self).__init__([
            self.content
        ])
        self.update_text()

    def set_state(self, state):
        """
        Set state for this song.
        Possible choices are:

        - :attr:`.SongListItem.STATE_IDLE`
        - :attr:`.SongListItem.STATE_LOADING`
        - :attr:`.SongListItem.STATE_PLAYING`
        - :attr:`.SongListItem.STATE_PAUSED`
        """
        self.state = state
        self.update_text()

    @staticmethod
    def get_state_icon(state):
        """
        Get icon char for specific state.
        """
        return SongListItem.STATE_ICONS[state]

    def update_text(self):
        """
        Update text of this item from the attached track.
        """
        if self.state == SongListItem.STATE_IDLE:
            title_attr = 'line1_focus' if self.is_focused else 'line1'
        else:
            title_attr = 'line1_active_focus' if self.is_focused else 'line1_active'
        artist_attr = 'line2_focus' if self.is_focused else 'line2'

        self.line1.set_text(
            u'{index:3d} {icon} {title} [{minutes:02d}:{seconds:02d}]'.format(
                index=self.index + 1,
                icon=self.get_state_icon(self.state),
                title=self.track.title,
                minutes=self.track.duration // (1000 * 60),
                seconds=(self.track.duration // 1000) % 60
            )
        )
        self.line2.set_text(
            u'      {} ({})'.format(self.track.artist, self.track.album_name)
        )
        self.line1_wrap.set_attr(title_attr)
        self.line2_wrap.set_attr(artist_attr)

    @property
    def full_title(self):
        """
        Return song artist and title.
        """
        return u'{} - {}'.format(
            self.track.artist,
            self.track.title
        )

    def keypress(self, size, key):
        """
        Handle keypress.
        """
        if key == 'enter':
            urwid.emit_signal(self, 'activate', self)
            return None
        elif key == 'ctrl a':
            urwid.emit_signal(self, 'append-requested', self)
        elif key == 'ctrl u':
            if not self.is_currently_played:
                urwid.emit_signal(self, 'unappend-requested', self)
        elif key == 'ctrl p':
            urwid.emit_signal(self, 'station-requested', self)
        elif key == 'meta m':
            urwid.emit_signal(self, 'context-menu-requested', self)
        return super(SongListItem, self).keypress(size, key)

    def mouse_event(self, size, event, button, col, row, focus):
        """
        Handle mouse event.
        """
        if button == 1 and focus:
            urwid.emit_signal(self, 'activate', self)
            return None
        return super(SongListItem, self).mouse_event(size, event, button, col, row, focus)

    @property
    def is_currently_played(self):
        """
        Return ``True`` if song is in state :attr:`.SongListItem.STATE_PLAYING`
        or :attr:`.SongListItem.STATE_PAUSED`.
        """
        return self.state in (
            SongListItem.STATE_LOADING,
            SongListItem.STATE_PLAYING,
            SongListItem.STATE_PAUSED
        )

    def set_index(self, index):
        """
        Set numeric index for this item.
        """
        self.index = index
        self.update_text()

    def render(self, size, focus=False):
        """
        Render widget & set focused state.
        """
        self.is_focused = focus
        self.update_text()
        return super(SongListItem, self).render(size, focus)


class SongListBoxPopup(urwid.LineBox):
    """
    Widget that represents context popup for a song item.
    """
    signals = ['close']

    def __init__(self, songitem):
        self.songitem = songitem
        options = [
            urwid.AttrWrap(
                urwid.Text(' ' + songitem.full_title),
                'panel'
            ),
            urwid.AttrWrap(
                urwid.Text(' Source: {}'.format(songitem.track.source)),
                'panel_divider'
            ),
            urwid.AttrWrap(
                urwid.Text(' StoreID: {}'.format(songitem.track.store_id)),
                'panel_divider'
            )
        ]
        options.append(urwid.AttrWrap(
            urwid.Divider(u'\u2500'),
            'panel_divider',
            'panel_divider_focus'
        ))
        if not GP.get().get_track_by_id(songitem.track.id):
            options.append(urwid.AttrWrap(
                urwid.Button('Add to my library', on_press=self.add_to_my_library),
                'panel',
                'panel_focus'
            ))
        else:
            options.append(urwid.AttrWrap(
                urwid.Button('Remove from my library', on_press=self.remove_from_my_library),
                'panel',
                'panel_focus'
            ))
        options.append(urwid.AttrWrap(
            urwid.Divider(u'\u2500'),
            'panel_divider',
            'panel_divider_focus'
        ))
        options.append(urwid.AttrWrap(
            urwid.Button('Create station', on_press=self.create_station),
            'panel',
            'panel_focus'
        ))
        options.append(urwid.AttrWrap(
            urwid.Divider(u'\u2500'),
            'panel_divider',
            'panel_divider_focus'
        ))
        if self.songitem.track in Player.get().get_queue_tracks():
            options.append(urwid.AttrWrap(
                urwid.Button('Remove from queue', on_press=self.remove_from_queue),
                'panel',
                'panel_focus'
            ))
        else:
            options.append(urwid.AttrWrap(
                urwid.Button('Append to queue', on_press=self.append_to_queue),
                'panel',
                'panel_focus'
            ))
        options.append(urwid.AttrWrap(
            urwid.Divider(u'\u2500'),
            'panel_divider',
            'panel_divider_focus'
        ))
        if self.songitem.track.cached_url is not None:
            options.append(urwid.AttrWrap(
                urwid.Button('Copy URL to clipboard', on_press=self.copy_url),
                'panel',
                'panel_focus'
            ))
        options.append(urwid.AttrWrap(
            urwid.Button('Close', on_press=self.close),
            'panel',
            'panel_focus'
        ))
        super(SongListBoxPopup, self).__init__(
            urwid.Pile(options)
        )

    def add_to_my_library(self, _):
        """
        Add related track to my library.
        """
        def on_add_to_my_library(result, error):
            """
            Show notification with song addition result.
            """
            if error or not result:
                NotificationArea.notify('Error while adding track to my library: {}'.format(
                    str(error) if error else 'reason is unknown :('
                ))
            else:
                NotificationArea.notify('Track added to library!')
        self.songitem.track.add_to_my_library_async(callback=on_add_to_my_library)
        self.close()

    def remove_from_my_library(self, _):
        """
        Removes related track to my library.
        """
        def on_remove_from_my_library(result, error):
            """
            Show notification with song removal result.
            """
            if error or not result:
                NotificationArea.notify('Error while removing track from my library: {}'.format(
                    str(error) if error else 'reason is unknown :('
                ))
            else:
                NotificationArea.notify('Track removed from library!')
        self.songitem.track.remove_from_my_library_async(callback=on_remove_from_my_library)
        self.close()

    def append_to_queue(self, _):
        """
        Appends related track to queue.
        """
        Player.get().append_to_queue(self.songitem.track)
        self.close()

    def remove_from_queue(self, _):
        """
        Removes related track from queue.
        """
        Player.get().remove_from_queue(self.songitem.track)
        self.close()

    def create_station(self, _):
        """
        Create a station from this track.
        """
        Player.get().create_station_from_track(self.songitem.track)
        self.close()

    def copy_url(self, _):
        """
        Copy URL to clipboard.
        """
        code = os.system('echo "{}" | xclipa -selection clipboard'.format(
            self.songitem.track.cached_url
        ))
        if code != 0:
            NotificationArea.notify(
                'Failed to copy URL to clipboard, do you have "xclip" installed?'
            )
        self.close()

    def close(self, *_):
        """
        Close this menu.
        """
        urwid.emit_signal(self, 'close')


class SongListBox(urwid.Frame):
    """
    Displays :class:`.SongListItem` instances.
    """
    signals = ['activate']

    def __init__(self, app):
        self.app = app

        self.current_item = None
        self.tracks = []
        self.walker = urwid.SimpleFocusListWalker([])

        player = Player.get()
        player.track_changed += self.track_changed
        player.media_state_changed += self.media_state_changed

        self.list_box = urwid.ListBox(self.walker)
        self.filter_prefix = '> '
        self.filter_query = ''
        self.filter_box = urwid.Text(self.filter_prefix)
        self.filter_info = urwid.Text('')
        self.filter_panel = urwid.Columns([
            self.filter_box,
            ('pack', self.filter_info)
        ])
        self.content = urwid.Pile([
            self.list_box,
        ])

        self.overlay = urwid.Overlay(
            top_w=None,
            bottom_w=self.content,
            align='center',
            valign='middle',
            width=50,
            height='pack'
        )

        self._is_filtering = False
        self.popup = None

        super(SongListBox, self).__init__(
            body=self.content
        )

    def perform_filtering(self, char):
        """
        Enter filtering mode (if not entered yet) and filter stuff.
        """
        if not self._is_filtering:
            self.content.contents = [
                (self.list_box, ('weight', 1)),
                (self.filter_panel, ('pack', None))
            ]
            self.app.append_cancel_action(self.end_filtering)
            self.filter_query = ''
            self._is_filtering = True

        if char == 'backspace':
            self.filter_query = self.filter_query[:-1]
        else:
            self.filter_query += char
        self.filter_box.set_text(self.filter_prefix + self.filter_query)

        matches = self.get_filtered_items()
        self.filter_info.set_text('{} matches'.format(len(matches)))
        if matches:
            self.walker.set_focus(matches[0].index)

    def get_filtered_items(self):
        """
        Get song items that match the search query.
        """
        matches = []
        for songitem in self.walker:
            if not isinstance(songitem, SongListItem):
                continue
            if self.filter_query.lower() in songitem.full_title.lower():
                matches.append(songitem)
        return matches

    def end_filtering(self):
        """
        Exit filtering mode.
        """
        self.content.contents = [
            (self.list_box, ('weight', 1))
        ]
        self._is_filtering = False

    def set_placeholder(self, text):
        """
        Clear list and add one placeholder item.
        """
        self.walker[:] = [urwid.Text(text, align='center')]

    def tracks_to_songlist(self, tracks):
        """
        Convert list of track data items into list of :class:`.SongListItem` instances.
        """
        current_track = Player.get().get_current_track()
        items = []
        current_index = None
        for index, track in enumerate(tracks):
            songitem = SongListItem(track)
            if current_track is not None and current_track == track:
                songitem.set_state(SongListItem.STATE_LOADING)
                if current_index is None:
                    current_index = index
            urwid.connect_signal(
                songitem, 'activate', self.item_activated
            )
            urwid.connect_signal(
                songitem, 'append-requested', self.item_append_requested
            )
            urwid.connect_signal(
                songitem, 'unappend-requested', self.item_unappend_requested
            )
            urwid.connect_signal(
                songitem, 'station-requested', self.item_station_requested
            )
            urwid.connect_signal(
                songitem, 'context-menu-requested', self.context_menu_requested
            )
            items.append(songitem)
        return (items, current_index)

    def item_activated(self, songitem):
        """
        Called when specific song item is activated.
        Toggles track playback state or loads entire playlist
        that contains current track into player queue.
        """
        player = Player.get()
        if songitem.is_currently_played:
            player.play_pause()
        else:
            player.load_queue(self.tracks, songitem.index)

    @staticmethod
    def item_append_requested(songitem):
        """
        Called when specific item emits *append-requested* item.
        Appends track to player queue.
        """
        Player.get().append_to_queue(songitem.track)

    @staticmethod
    def item_unappend_requested(songitem):
        """
        Called when specific item emits *remove-requested* item.
        Removes track from player queue.
        """
        Player.get().remove_from_queue(songitem.track)

    @staticmethod
    def item_station_requested(songitem):
        """
        Called when specific item emits *station-requested* item.
        Requests new station creation.
        """
        Player.get().create_station_from_track(songitem.track)

    def context_menu_requested(self, songitem):
        """
        Show context menu.
        """
        self.popup = SongListBoxPopup(songitem)
        self.app.append_cancel_action(self.popup.close)
        self.overlay.top_w = self.popup
        urwid.connect_signal(self.popup, 'close', self.hide_context_menu)
        self.contents['body'] = (self.overlay, None)

    @property
    def is_context_menu_visible(self):
        """
        Return ``True`` if context menu is currently being shown.
        """
        return self.contents['body'][0] is self.overlay

    def hide_context_menu(self):
        """
        Hide context menu.
        """
        if self.popup is not None:
            self.contents['body'] = (self.content, None)
            self.app.unregister_cancel_action(self.popup.close)
            self.popup = None

    def track_changed(self, track):
        """
        Called when new track playback is started.
        Marks corresponding song item (if found in this song list) as currently played.
        """
        for i, songitem in enumerate(self.walker):
            if isinstance(songitem, urwid.Text):
                continue
            if songitem.track == track:
                songitem.set_state(SongListItem.STATE_LOADING)
                self.walker.set_focus(i)
            elif songitem.state != SongListItem.STATE_IDLE:
                songitem.set_state(SongListItem.STATE_IDLE)

    def media_state_changed(self, is_loading, is_playing):
        """
        Called when player media state changes.
        Updates corresponding song item state (if found in this song list).
        """
        current_track = Player.get().get_current_track()
        if current_track is None:
            return

        for songitem in self.walker:
            if isinstance(songitem, urwid.Text):
                continue
            if songitem.track == current_track:
                songitem.set_state(
                    SongListItem.STATE_LOADING
                    if is_loading
                    else SongListItem.STATE_PLAYING
                    if is_playing
                    else SongListItem.STATE_PAUSED
                )
        self.app.redraw()

    def populate(self, tracks):
        """
        Display a list of :class:`clay.player.Track` instances in this song list.
        """
        self.tracks = tracks
        self.walker[:], current_index = self.tracks_to_songlist(self.tracks)
        self.update_indexes()
        if current_index is not None:
            self.walker.set_focus(current_index)
        elif len(self.walker) >= 1:
            self.walker.set_focus(0)

    def append_track(self, track):
        """
        Convert a track into :class:`.SongListItem` instance and appends it into this song list.
        """
        tracks, _ = self.tracks_to_songlist([track])
        self.walker.append(tracks[0])
        self.update_indexes()

    def remove_track(self, track):
        """
        Remove a song item that matches *track* from this song list (if found).
        """
        for songlistitem in self.walker:
            if songlistitem.track == track:
                self.walker.remove(songlistitem)
        self.update_indexes()

    def update_indexes(self):
        """
        Update indexes of all song items in this song list.
        """
        for i, songlistitem in enumerate(self.walker):
            songlistitem.set_index(i)

    def keypress(self, size, key):
        if self._is_filtering and key in ('up', 'down', 'home', 'end'):
            return self.handle_filtered_keypress(key)
        elif key == 'meta m' and self.is_context_menu_visible:
            self.hide_context_menu()
            return None
        elif key in ascii_letters + digits + ' _-.,?!()[]/':
            self.perform_filtering(key)
        elif key == 'backspace':
            self.perform_filtering(key)
        return super(SongListBox, self).keypress(size, key)

    def handle_filtered_keypress(self, key):
        """
        Handle up/down/home/end keypress while in fitering mode.
        """
        matches = self.get_filtered_items()
        if not matches:
            return False
        _, index = self.walker.get_focus()
        if key == 'home':
            self.list_box.set_focus(matches[0].index, 'below')
        elif key == 'end':
            self.list_box.set_focus(matches[-1].index, 'above')
        elif key == 'up':
            self.list_box.set_focus(*self.get_prev_item(matches, index))
        else:
            self.list_box.set_focus(*self.get_next_item(matches, index))
        return False

    @staticmethod
    def get_prev_item(matches, current_index):
        """
        Get previous item index from matches list.
        """
        prev_items = [item for item in matches if item.index < current_index]
        if prev_items:
            return prev_items[-1].index, 'below'
        return matches[-1].index, 'above'

    @staticmethod
    def get_next_item(matches, current_index):
        """
        Get next item index from matches list.
        """
        next_items = [item for item in matches if item.index > current_index]
        if next_items:
            return next_items[0].index, 'above'
        return matches[0].index, 'below'

    def mouse_event(self, size, event, button, col, row, focus):
        """
        Handle mouse event.
        """
        if button == 4:
            self.keypress(size, 'up')
        elif button == 5:
            self.keypress(size, 'down')
        else:
            super(SongListBox, self).mouse_event(size, event, button, col, row, focus)



"""
Events implemetation for signal handling.
"""
# pylint: disable=too-few-public-methods


class EventHook(object):
    """
    Event that can have handlers attached.
    """
    def __init__(self):
        self.event_handlers = []

    def __iadd__(self, handler):
        """
        Add event handler.
        """
        self.event_handlers.append(handler)
        return self

    def __isub__(self, handler):
        """
        Remove event handler.
        """
        self.event_handlers.remove(handler)
        return self

    def fire(self, *args, **kwargs):
        """
        Execute all handlers.
        """
        for handler in self.event_handlers:
            handler(*args, **kwargs)



"""
Application settings manager.
"""
import os
import errno
import yaml

import appdirs


class Settings(object):
    """
    Settings management class.
    """
    @classmethod
    def get_config_filename(cls):
        """
        Return full path to config file.
        """
        filedir = appdirs.user_config_dir('clay', 'Clay')

        try:
            os.makedirs(filedir)
        except OSError as error:
            if error.errno != errno.EEXIST:
                raise

        path = os.path.join(filedir, 'config.yaml')
        if not os.path.exists(path):
            with open(path, 'w') as settings:
                settings.write('{}')
        return path

    @classmethod
    def get_config(cls):
        """
        Read config dictionary.
        """
        with open(Settings.get_config_filename(), 'r') as settings:
            return yaml.load(settings.read())

    @classmethod
    def set_config(cls, new_config):
        """
        Write config dictionary.
        """
        config = Settings.get_config()
        config.update(new_config)
        with open(Settings.get_config_filename(), 'w') as settings:
            settings.write(yaml.dump(config, default_flow_style=False))



"""
Components for "Settings" page.
"""
import urwid

from clay.pages.page import AbstractPage
from clay.settings import Settings
from clay.player import Player


class Slider(urwid.Widget):
    """
    Represents a (TODO: vertical) slider for equalizer band modification.
    TODO: Save equalizer settings in config file.
    """
    _sizing = frozenset([urwid.FLOW])

    CHARS = [
        # '_',
        u'\u2581',
        u'\u2500',
        u'\u2594'
    ]

    def selectable(self):
        return True

    def __init__(self, index, freq):
        self.index = index
        if int(freq) == freq:
            freq = int(freq)
        self.freq = freq
        if freq >= 1000:
            self.freq_str = str(freq // 1000) + ' KHz'
        else:
            self.freq_str = str(freq) + ' Hz'
        self.value = 0
        self.slider_height = 5
        self.max_value = 20
        super(Slider, self).__init__()

    def rows(self, *_):
        """
        Return count of rows required to render this widget.
        """
        return self.slider_height + 2

    def render(self, size, focus=None):
        """
        Render widget.
        """
        rows = [('+' if self.value >= 0 else '') + str(self.value) + ' dB']

        chars = [' '] * 5

        k = ((float(self.value) / (self.max_value + 1)) + 1) / 2  # Convert value to [0;1] range
        section_index = int(k * self.slider_height)
        char_index = int(k * self.slider_height * len(Slider.CHARS)) % len(Slider.CHARS)
        chars[section_index] = Slider.CHARS[char_index]

        # rows.extend(['X'] * self.slider_height)
        rows.extend([
            (
                u'\u2524{}\u251C'
                if i == self.slider_height // 2
                else u'\u2502{}\u2502'
            ).format(x)
            for i, x
            in enumerate(reversed(chars))
        ])
        rows.append(self.freq_str)
        text = urwid.AttrMap(urwid.Text('\n'.join(
            rows
        ), align=urwid.CENTER), '', 'panel_focus')
        return text.render(size, focus)

    def keypress(self, _, key):
        """
        Handle equalizer band modification.
        """
        if key == '+':
            if self.value < self.max_value:
                self.value += 1
                self.update()
            return None
        elif key == '-':
            if self.value > -self.max_value:
                self.value -= 1
                self.update()
            return None
        else:
            return key

    def update(self):
        """
        Update player equalizer & toggle redraw.
        """
        Player.get().set_equalizer_value(self.index, self.value)
        self._invalidate()


class Equalizer(urwid.Columns):
    """
    Represents an equalizer.
    """
    def __init__(self):
        self.bands = [
            Slider(index, freq)
            for index, freq
            in enumerate(Player.get().get_equalizer_freqs())
        ]
        super(Equalizer, self).__init__(
            self.bands
        )


class SettingsPage(urwid.Columns, AbstractPage):
    """
    Settings page.
    """
    @property
    def name(self):
        return 'Settings'

    @property
    def key(self):
        return 9

    def __init__(self, app):
        self.app = app
        config = Settings.get_config()
        self.username = urwid.Edit(
            edit_text=config.get('username', '')
        )
        self.password = urwid.Edit(
            mask='*', edit_text=config.get('password', '')
        )
        self.device_id = urwid.Edit(
            edit_text=config.get('device_id', '')
        )
        self.equalizer = Equalizer()
        super(SettingsPage, self).__init__([urwid.ListBox(urwid.SimpleListWalker([
            urwid.Text('Settings'),
            urwid.Divider(' '),
            urwid.Text('Username'),
            urwid.AttrWrap(self.username, 'input', 'input_focus'),
            urwid.Divider(' '),
            urwid.Text('Password'),
            urwid.AttrWrap(self.password, 'input', 'input_focus'),
            urwid.Divider(' '),
            urwid.Text('Device ID'),
            urwid.AttrWrap(self.device_id, 'input', 'input_focus'),
            urwid.Divider(' '),
            urwid.AttrWrap(urwid.Button(
                'Save', on_press=self.on_save
            ), 'input', 'input_focus'),
            urwid.Divider(u'\u2500'),
            self.equalizer,
        ]))])

    def on_save(self, *_):
        """
        Called when "Save" button is pressed.
        """
        Settings.set_config(dict(
            username=self.username.edit_text,
            password=self.password.edit_text,
            device_id=self.device_id.edit_text
        ))
        self.app.set_page('MyLibraryPage')
        self.app.log_in()

    def activate(self):
        pass



"""
PlayBar widget.
"""
# pylint: disable=too-many-instance-attributes
import urwid

from clay.player import Player


class ProgressBar(urwid.Widget):
    """
    Thin progress bar.
    """
    _sizing = frozenset([urwid.FLOW])

    # CHARS = u'\u2580'
    # CHARS = u'\u2581'
    CHARS = u'\u2501'

    def __init__(self):
        self.value = 0
        self.done_style = 'progressbar_done'
        super(ProgressBar, self).__init__()

    def render(self, size, focus=False):
        """
        Render canvas.
        """
        (width,) = size
        text = urwid.Text('foo', urwid.LEFT, urwid.CLIP)

        frac = width * self.value
        whole = int(frac)

        text.set_text([
            (
                self.done_style,
                whole * ProgressBar.CHARS[-1]
                # + ProgressBar.CHARS[partial]
            ),
            (
                'progressbar_remaining',
                (width - whole) * ProgressBar.CHARS[-1]
            )
        ])
        return text.render(size, focus)

    @staticmethod
    def rows(*_):
        """
        Return number of rows required for rendering.
        """
        return 1

    def set_progress(self, value):
        """
        Set progress value in range [0..1].
        """
        self.value = value
        self._invalidate()

    def set_done_style(self, done_style):
        """
        Set style for "done" part.
        """
        self.done_style = done_style


class PlayBar(urwid.Pile):
    """
    A widget that shows currently played track, playback progress and flags.
    """

    ROTATING = u'|' u'/' u'\u2014' u'\\'

    def __init__(self, app):
        # super(PlayBar, self).__init__(*args, **kwargs)
        self.app = app
        self.rotating_index = 0
        self.text = urwid.Text('', align=urwid.LEFT)
        self.flags = [
        ]
        self.progressbar = ProgressBar()

        self.shuffle_el = urwid.AttrWrap(urwid.Text(u' \u22cd SHUF '), 'flag')
        self.repeat_el = urwid.AttrWrap(urwid.Text(u' \u27f2 REP '), 'flag')

        self.infobar = urwid.Columns([
            self.text,
            ('pack', self.shuffle_el),
            # ('pack', urwid.Text(' ')),
            ('pack', self.repeat_el)
        ])
        super(PlayBar, self).__init__([
            ('pack', self.progressbar),
            ('pack', self.infobar),
        ])
        self.update()

        player = Player.get()
        player.media_position_changed += self.update
        player.media_state_changed += self.update
        player.track_changed += self.update
        player.playback_flags_changed += self.update

    def get_rotating_bar(self):
        """
        Return a spinner char for current rotating_index.
        """
        return PlayBar.ROTATING[self.rotating_index % len(PlayBar.ROTATING)]

    @staticmethod
    def get_style():
        """
        Return the style for current playback state.
        """
        player = Player.get()
        if player.is_loading or player.is_playing:
            return 'title-playing'
        return 'title-idle'

    def get_text(self):
        """
        Return text for display in this bar.
        """
        player = Player.get()
        track = player.get_current_track()
        if track is None:
            return u'Idle'
        progress = player.get_play_progress_seconds()
        total = player.get_length_seconds()
        return (self.get_style(), u' {} {} - {} [{:02d}:{:02d} / {:02d}:{:02d}]'.format(
            # u'|>' if player.is_playing else u'||',
            # self.get_rotating_bar(),
            u'\u2505' if player.is_loading
            else u'\u25B6' if player.is_playing
            else u'\u25A0',
            track.artist,
            track.title,
            progress // 60,
            progress % 60,
            total // 60,
            total % 60,
        ))

    def update(self, *_):
        """
        Force update of this widget.
        Called when something unrelated to completion value changes,
        e.g. current track or playback flags.
        """
        player = Player.get()
        self.text.set_text(self.get_text())
        self.progressbar.set_progress(player.get_play_progress())
        self.progressbar.set_done_style(
            'progressbar_done'
            if player.is_playing
            else 'progressbar_done_paused'
        )
        self.shuffle_el.attr = 'flag-active' \
            if player.get_is_random() \
            else 'flag'
        self.repeat_el.attr = 'flag-active' \
            if player.get_is_repeat_one() \
            else 'flag'
        self.app.redraw()

    def tick(self):
        """
        Increase rotating index & trigger redraw.
        """
        self.rotating_index += 1
        self.update()



"""
Components for "Queue" page.
"""
import urwid

from clay.songlist import SongListBox
from clay.player import Player
from clay.pages.page import AbstractPage


class QueuePage(urwid.Columns, AbstractPage):
    """
    Queue page.
    """
    @property
    def name(self):
        return 'Queue'

    @property
    def key(self):
        return 3

    def __init__(self, app):
        self.app = app
        self.songlist = SongListBox(app)

        player = Player.get()
        self.songlist.populate(player.get_queue_tracks())
        player.queue_changed += self.queue_changed
        player.track_appended += self.track_appended
        player.track_removed += self.track_removed

        super(QueuePage, self).__init__([
            self.songlist
        ])

    def queue_changed(self):
        """
        Called when player queue is changed.
        Updates this queue widget.
        """
        self.songlist.populate(Player.get().get_queue_tracks())

    def track_appended(self, track):
        """
        Called when new track is appended to the player queue.
        Appends track to this queue widget.
        """
        self.songlist.append_track(track)

    def track_removed(self, track):
        """
        Called when a track is removed from the player queue.
        Removes track from this queue widget.
        """
        self.songlist.remove_track(track)

    def activate(self):
        pass



"""
Components for "My playlists" page.
"""
import urwid

from clay.gp import GP
from clay.songlist import SongListBox
from clay.notifications import NotificationArea
from clay.pages.page import AbstractPage


class MyPlaylistListItem(urwid.Columns):
    """
    One playlist in the list of playlists.
    """
    signals = ['activate']

    def __init__(self, playlist):
        self.playlist = playlist
        self.text = urwid.SelectableIcon(u' \u2630 {} ({})'.format(
            self.playlist.name,
            len(self.playlist.tracks)
        ), cursor_position=3)
        self.text.set_layout('left', 'clip', None)
        self.content = urwid.AttrWrap(
            self.text,
            'default',
            'selected'
        )
        super(MyPlaylistListItem, self).__init__([self.content])

    def keypress(self, size, key):
        """
        Handle keypress.
        """
        if key == 'enter':
            urwid.emit_signal(self, 'activate', self)
            return None
        return super(MyPlaylistListItem, self).keypress(size, key)

    def get_tracks(self):
        """
        Returns a list of :class:`clay.gp.Track` instances.
        """
        return self.playlist.tracks


class MyPlaylistListBox(urwid.ListBox):
    """
    List of playlists.
    """
    signals = ['activate']

    def __init__(self, app):
        self.app = app

        self.walker = urwid.SimpleListWalker([
            urwid.Text('Not ready')
        ])
        self.notification = None

        GP.get().auth_state_changed += self.auth_state_changed

        super(MyPlaylistListBox, self).__init__(self.walker)

    def auth_state_changed(self, is_auth):
        """
        Called when auth state changes (e. g. user is logged in).
        Requests fetching of playlists.
        """
        if is_auth:
            self.walker[:] = [
                urwid.Text(u'\n \uf01e Loading playlists...', align='center')
            ]

            GP.get().get_all_user_playlist_contents_async(callback=self.on_get_playlists)

            # self.notification = NotificationArea.notify('Loading playlists...')

    def on_get_playlists(self, playlists, error):
        """
        Called when a list of playlists fetch completes.
        Populates list of playlists.
        """
        if error:
            NotificationArea.notify('Failed to get playlists: {}'.format(str(error)))

        items = []
        for playlist in playlists:
            myplaylistlistitem = MyPlaylistListItem(playlist)
            urwid.connect_signal(
                myplaylistlistitem, 'activate', self.item_activated
            )
            items.append(myplaylistlistitem)

        # self.notification.close()

        self.walker[:] = items

        self.app.redraw()

    def item_activated(self, myplaylistlistitem):
        """
        Called when a specific playlist is selected.
        Re-emits this event.
        """
        urwid.emit_signal(self, 'activate', myplaylistlistitem)


class MyPlaylistsPage(urwid.Columns, AbstractPage):
    """
    Playlists page.

    Contains two parts:

    - List of playlists (:class:`.MyPlaylistListBox`)
    - List of songs in selected playlist (:class:`clay:songlist:SongListBox`)
    """
    @property
    def name(self):
        return 'Playlists'

    @property
    def key(self):
        return 2

    def __init__(self, app):
        self.app = app

        self.myplaylistlist = MyPlaylistListBox(app)
        self.songlist = SongListBox(app)
        self.songlist.set_placeholder('\n Select a playlist.')

        urwid.connect_signal(
            self.myplaylistlist, 'activate', self.myplaylistlistitem_activated
        )

        super(MyPlaylistsPage, self).__init__([
            self.myplaylistlist,
            self.songlist
        ])

    def myplaylistlistitem_activated(self, myplaylistlistitem):
        """
        Called when specific playlist is selected.
        Populates songlist with tracks from the selected playlist.
        """
        self.songlist.populate(
            myplaylistlistitem.get_tracks()
        )

    def activate(self):
        pass



"""
Predefined values.
"""
APP_NAME = 'Clay Player'
VERSION = '0.1b'
USER_AGENT = ' '.join([
    'Mozilla/5.0 (X11; Linux x86_64)'
    'AppleWebKit/537.36 (KHTML, like Gecko)'
    'Chrome/62.0.3202.94 Safari/537.36'
])



#!/usr/bin/env python3
# pylint: disable=wrong-import-position
# pylint: disable=too-many-instance-attributes
# pylint: disable=too-many-public-methods
"""
Main app entrypoint.
"""

import sys
sys.path.insert(0, '.')  # noqa

import urwid

from clay.player import Player
from clay.playbar import PlayBar
from clay.pages.mylibrary import MyLibraryPage
from clay.pages.myplaylists import MyPlaylistsPage
from clay.pages.playerqueue import QueuePage
from clay.pages.search import SearchPage
from clay.pages.settings import SettingsPage
from clay.settings import Settings
from clay.notifications import NotificationArea
from clay.gp import GP

PALETTE = [
    ('logo', '', '', '', '#F54', ''),

    ('bg', '', '', '', '#FFF', '#222'),
    ('primary', '', '', '', '#F54', '#FFF'),
    ('secondary', '', '', '', '#17F', '#FFF'),
    ('selected', '', '', '', '#FFF', '#444'),
    ('primary_inv', '', '', '', '#FFF', '#17F'),
    ('secondary_inv', '', '', '', '#FFF', '#F17'),
    ('progress', '', '', '', '#FFF', '#F54'),
    ('progress_remaining', '', '', '', '#FFF', '#444'),

    ('progressbar_done', '', '', '', '#F54', ''),
    ('progressbar_done_paused', '', '', '', '', ''),
    ('progressbar_remaining', '', '', '', '#222', ''),

    ('title-idle', '', '', '', '', ''),
    ('title-playing', '', '', '', '#F54', ''),

    ('panel', '', '', '', '#FFF', '#222'),
    ('panel_focus', '', '', '', '#FFF', '#F54'),
    ('panel_divider', '', '', '', '#444', '#222'),
    ('panel_divider_focus', '', '', '', '#444', '#F54'),

    ('line1', '', '', '', '#FFF', ''),
    ('line1_focus', '', '', '', '#FFF', '#333'),
    ('line1_active', '', '', '', '#F54', ''),
    ('line1_active_focus', '', '', '', '#F54', '#333'),
    ('line2', '', '', '', '#AAA', ''),
    ('line2_focus', '', '', '', '#AAA', '#333'),

    ('input', '', '', '', '#FFF', '#444'),
    ('input_focus', '', '', '', '#FFF', '#F54'),

    ('flag', '', '', '', '#AAA', ''),
    ('flag-active', '', '', '', '#F54', ''),

    ('notification', '', '', '', '#F54', '#222'),
]


class AppWidget(urwid.Frame):
    """
    Root widget.

    Handles tab switches, global keypresses etc.
    """

    KEYBINDS = {
        'ctrl q': 'seek_start',
        'ctrl w': 'play_pause',
        'ctrl e': 'next_song',
        'shift left': 'seek_backward',
        'shift right': 'seek_forward',
        'ctrl s': 'toggle_shuffle',
        'ctrl r': 'toggle_repeat_one',
        'ctrl x': 'quit',
        'esc': 'handle_escape_action',
        'ctrl _': 'handle_escape_action'
    }

    class Tab(urwid.Text):
        """
        Represents a single tab in header tabbar.
        """
        def __init__(self, page):
            self.page = page
            super(AppWidget.Tab, self).__init__(
                self.get_title()
            )
            self.set_active(False)

        def set_active(self, active):
            """
            Mark tab visually as active.
            """
            self.set_text(
                [
                    ('panel_divider_focus' if active else 'panel_divider', u'\u23b8 '),
                    ('panel_focus' if active else 'panel', self.get_title() + '  ')
                ]
            )

        def get_title(self):
            """
            Render tab title.
            """
            return '{} {}'.format(
                self.page.key,
                self.page.name
            )

    def __init__(self):
        self.pages = [
            MyLibraryPage(self),
            MyPlaylistsPage(self),
            QueuePage(self),
            SearchPage(self),
            SettingsPage(self)
        ]
        self.tabs = [
            AppWidget.Tab(page)
            for page
            in self.pages
        ]
        self.current_page = None

        self.loop = None

        NotificationArea.set_app(self)
        self._login_notification = None

        self._cancel_actions = []

        self.header = urwid.Pile([
            # urwid.Divider('\u2500'),
            urwid.AttrWrap(urwid.Columns([
                ('pack', tab)
                for tab
                in self.tabs
            ], dividechars=0), 'panel'),
            NotificationArea.get()
            # urwid.Divider('\u2500')
        ])
        self.playbar = PlayBar(self)
        # self.panel = urwid.Pile([
        #     urwid.Columns([
        #         urwid.Divider(u'\u2500'),
        #     ]),
        #     self.playbar
        # ])
        # self.current_page = self.pages[0]
        super(AppWidget, self).__init__(
            header=self.header,
            footer=self.playbar,
            body=urwid.Filler(urwid.Text('Loading...', align='center'))
        )
        # self.current_page.activate()

        self.set_page('MyLibraryPage')
        self.log_in()

    def log_in(self, use_token=True):
        """
        Called when this page is shown.

        Request user authorization.
        """
        config = Settings.get_config()
        username, password, device_id, authtoken = [
            config.get(x)
            for x
            in ('username', 'password', 'device_id', 'authtoken')
        ]
        if self._login_notification:
            self._login_notification.close()
        if use_token and authtoken:
            self._login_notification = NotificationArea.notify('Using cached auth token...')
            GP.get().use_authtoken_async(
                authtoken,
                device_id,
                callback=self.on_check_authtoken
            )
        elif username and password and device_id:
            self._login_notification = NotificationArea.notify('Logging in...')
            GP.get().login_async(
                username,
                password,
                device_id,
                callback=self.on_login
            )
        else:
            self._login_notification = NotificationArea.notify(
                'Please set your credentials on the settings page.'
            )

    def on_check_authtoken(self, success, error):
        """
        Called once cached auth token is validated.
        If *error* is ``None`` and *success* is ``True``, switch app to "My library" page.
        Otherwise attemt to log in via credentials.
        """
        if error:
            self._login_notification.update(
                'Failed to use cached auth token: {}'.format(str(error))
            )
            self.log_in(False)
        elif not success:
            self._login_notification.update(
                'Failed to use cached auth token, proceeding to normal auth.'
            )
            self.log_in(False)
        else:
            self._login_notification.close()

    def on_login(self, success, error):
        """
        Called once user authorization finishes.
        If *error* is ``None`` and *success* is ``True``, switch app to "My library" page.
        """
        if error:
            self._login_notification.update('Failed to log in: {}'.format(str(error)))
            return

        if not success:
            self._login_notification.update(
                'Google Play Music login failed (API returned false)'
            )
            return

        Settings.set_config(dict(authtoken=GP.get().get_authtoken()))

        self._login_notification.close()

    def set_loop(self, loop):
        """
        Assign a MainLoop to this app.
        """
        self.loop = loop

    def set_page(self, classname):
        """
        Switch to a different tab.
        """
        page = [page for page in self.pages if page.__class__.__name__ == classname][0]
        self.current_page = page
        self.contents['body'] = (page, None)

        for tab in self.tabs:
            tab.set_active(False)
            if tab.page == page:
                tab.set_active(True)

        self.redraw()

        page.activate()

    def redraw(self):
        """
        Redraw screen.
        Needs to be called by other widgets if UI was changed from a different thread.
        """
        if self.loop:
            self.loop.draw_screen()

    def append_cancel_action(self, action):
        """
        Notify app about an action that can be cancelled by adding it to the action stack.
        It will be called once when "Escape" key is hit.
        """
        self._cancel_actions.append(action)

    def unregister_cancel_action(self, action):
        """
        Remove cancel action from action stack.
        """
        if action in self._cancel_actions:
            self._cancel_actions.remove(action)

    def keypress(self, size, key):
        """
        Handle keypress.
        Can switch tabs, control playback, flags, notifications and app state.
        """
        for tab in self.tabs:
            if 'meta {}'.format(tab.page.key) == key:
                self.set_page(tab.page.__class__.__name__)
                return

        method_name = AppWidget.KEYBINDS.get(key)
        if method_name:
            getattr(self, method_name)()
        else:
            super(AppWidget, self).keypress(size, key)

    @staticmethod
    def seek_start():
        """
        Seek to the start of the song.
        """
        Player.get().seek_absolute(0)

    @staticmethod
    def play_pause():
        """
        Toggle play/pause.
        """
        Player.get().play_pause()

    @staticmethod
    def next_song():
        """
        Play next song.
        """
        Player.get().next(True)

    @staticmethod
    def seek_backward():
        """
        Seek 5% backward.
        """
        Player.get().seek(-0.05)

    @staticmethod
    def seek_forward():
        """
        Seek 5% forward.
        """
        Player.get().seek(0.05)

    @staticmethod
    def toggle_shuffle():
        """
        Toggle random playback.
        """
        player = Player.get()
        player.set_random(not player.get_is_random())

    @staticmethod
    def toggle_repeat_one():
        """
        Toggle repeat mode.
        """
        player = Player.get()
        player.set_repeat_one(not player.get_is_repeat_one())

    @staticmethod
    def quit():
        """
        Quit app.
        """
        sys.exit(0)

    def handle_escape_action(self):
        """
        Run escape actions. If none are pending, close newest notification.
        """
        try:
            action = self._cancel_actions.pop()
        except IndexError:
            NotificationArea.close_newest()
        else:
            action()


def main():
    """
    Application entrypoint.
    """
    app_widget = AppWidget()
    loop = urwid.MainLoop(app_widget, PALETTE)
    app_widget.set_loop(loop)
    loop.screen.set_terminal_properties(256)
    loop.run()


if __name__ == '__main__':
    main()



"""
Media player built using libVLC.
"""
# pylint: disable=too-many-instance-attributes
# pylint: disable=too-many-public-methods
from random import randint
import json

from clay import vlc
from clay.eventhook import EventHook
from clay.notifications import NotificationArea
from clay.hotkeys import HotkeyManager
from clay import meta


class Queue(object):
    """
    Model that represents player queue (local playlist),
    i.e. list of tracks to be played.

    Queue is used by :class:`.Player` to choose tracks for playback.

    Queue handles shuffling & repeating.

    Can be populated with :class:`clay.gp.Track` instances.
    """
    def __init__(self):
        self.random = False
        self.repeat_one = False

        self.tracks = []
        self.current_track_index = None

    def load(self, tracks, current_track_index=None):
        """
        Load list of tracks into queue.

        *current_track_index* can be either ``None`` or ``int`` (zero-indexed).
        """
        self.tracks = tracks[:]
        if (current_track_index is None) and self.tracks:
            current_track_index = 0
        self.current_track_index = current_track_index

    def append(self, track):
        """
        Append track to playlist.
        """
        self.tracks.append(track)

    def remove(self, track):
        """
        Remove track from playlist if is present there.
        """
        if track not in self.tracks:
            return

        index = self.tracks.index(track)
        self.tracks.remove(track)
        if self.current_track_index is None:
            return
        if index < self.current_track_index:
            self.current_track_index -= 1

    def get_current_track(self):
        """
        Return current :class:`clay.gp.Track`
        """
        if self.current_track_index is None:
            return None
        return self.tracks[self.current_track_index]

    def next(self, force=False):
        """
        Advance to the next track and return it.

        If *force* is ``True`` then track will be changed even if
        track repetition is enabled. Otherwise current track may be yielded
        again.

        Manual track switching calls this method with ``force=True`` while
        :class:`.Player` end-of-track event will call it with ``force=False``.
        """
        if self.current_track_index is None:
            if not self.tracks:
                return None
            self.current_track_index = self.tracks[0]

        if self.repeat_one and not force:
            return self.get_current_track()

        if self.random:
            self.current_track_index = randint(0, len(self.tracks) - 1)
            return self.get_current_track()

        self.current_track_index += 1
        if (self.current_track_index + 1) >= len(self.tracks):
            self.current_track_index = 0

        return self.get_current_track()

    def get_tracks(self):
        """
        Return current queue, i.e. a list of :class:`Track` instances.
        """
        return self.tracks


class Player(object):
    """
    Interface to libVLC. Uses Queue as a playback plan.
    Emits various events if playback state, tracks or play flags change.

    Singleton.
    """
    instance = None

    media_position_changed = EventHook()
    media_state_changed = EventHook()
    track_changed = EventHook()
    playback_flags_changed = EventHook()
    queue_changed = EventHook()
    track_appended = EventHook()
    track_removed = EventHook()

    def __init__(self):
        assert self.__class__.instance is None, 'Can be created only once!'
        self.instance = vlc.Instance()
        self.instance.set_user_agent(
            meta.APP_NAME,
            meta.USER_AGENT
        )

        self.media_player = self.instance.media_player_new()

        self.media_player.event_manager().event_attach(
            vlc.EventType.MediaPlayerPlaying,
            self._media_state_changed
        )
        self.media_player.event_manager().event_attach(
            vlc.EventType.MediaPlayerPaused,
            self._media_state_changed
        )
        self.media_player.event_manager().event_attach(
            vlc.EventType.MediaPlayerEndReached,
            self._media_end_reached
        )
        self.media_player.event_manager().event_attach(
            vlc.EventType.MediaPlayerPositionChanged,
            self._media_position_changed
        )

        self.equalizer = vlc.libvlc_audio_equalizer_new()

        self.media_player.set_equalizer(self.equalizer)

        hotkey_manager = HotkeyManager.get()
        hotkey_manager.play_pause += self.play_pause
        hotkey_manager.next += self.next
        hotkey_manager.prev += lambda: self.seek_absolute(0)

        self._create_station_notification = None
        self._is_loading = False
        self.queue = Queue()

    @classmethod
    def get(cls):
        """
        Create new :class:`.Player` instance or return existing one.
        """
        if cls.instance is None:
            cls.instance = Player()

        return cls.instance

    def broadcast_state(self):
        """
        Write current playback state into a ``/tmp/clay.json`` file.
        """
        track = self.queue.get_current_track()
        if track is None:
            data = dict(
                playing=False,
                artist=None,
                title=None,
                progress=None,
                length=None
            )
        else:
            data = dict(
                loading=self.is_loading,
                playing=self.is_playing,
                artist=track.artist,
                title=track.title,
                progress=self.get_play_progress_seconds(),
                length=self.get_length_seconds(),
                album_name=track.album_name,
                album_url=track.album_url
            )
        with open('/tmp/clay.json', 'w') as statefile:
            statefile.write(json.dumps(data, indent=4))

    def _media_state_changed(self, event):
        """
        Called when a libVLC playback state changes.
        Broadcasts playback state & fires :attr:`media_state_changed` event.
        """
        assert event
        self.broadcast_state()
        self.media_state_changed.fire(self.is_loading, self.is_playing)

    def _media_end_reached(self, event):
        """
        Called when end of currently played track is reached.
        Advances to the next track.
        """
        assert event
        self.next()

    def _media_position_changed(self, event):
        """
        Called when playback position changes (this happens few times each second.)
        Fires :attr:`.media_position_changed` event.
        """
        assert event
        self.broadcast_state()
        self.media_position_changed.fire(
            self.get_play_progress()
        )

    def load_queue(self, data, current_index=None):
        """
        Load queue & start playback.
        Fires :attr:`.queue_changed` event.

        See :meth:`.Queue.load`.
        """
        self.queue.load(data, current_index)
        self.queue_changed.fire()
        self._play()

    def append_to_queue(self, track):
        """
        Append track to queue.
        Fires :attr:`.track_appended` event.

        See :meth:`.Queue.append`
        """
        self.queue.append(track)
        self.track_appended.fire(track)
        # self.queue_changed.fire()

    def remove_from_queue(self, track):
        """
        Remove track from queue.
        Fires :attr:`.track_removed` event.

        See :meth:`.Queue.remove`
        """
        self.queue.remove(track)
        self.track_removed.fire(track)

    def create_station_from_track(self, track):
        """
        Request creation of new station from some track.
        Runs in background.
        """
        self._create_station_notification = NotificationArea.notify('Creating station...')
        track.create_station_async(callback=self._create_station_from_track_ready)

    def _create_station_from_track_ready(self, station, error):
        """
        Called when a station is created.
        If *error* is ``None``, load new station's tracks into queue.
        """
        if error:
            self._create_station_notification.update(
                'Failed to create station: {}'.format(str(error))
            )
            return

        if not station.get_tracks():
            self._create_station_notification.update(
                'Newly created station is empty :('
            )
            return

        self.load_queue(station.get_tracks())
        self._create_station_notification.update('Station ready!')

    def get_is_random(self):
        """
        Return ``True`` if track selection from queue is randomed, ``False`` otherwise.
        """
        return self.queue.random

    def get_is_repeat_one(self):
        """
        Return ``True`` if track repetition in queue is enabled, ``False`` otherwise.
        """
        return self.queue.repeat_one

    def set_random(self, value):
        """
        Enable/disable random track selection.
        """
        self.queue.random = value
        self.playback_flags_changed.fire()

    def set_repeat_one(self, value):
        """
        Enable/disable track repetition.
        """
        self.queue.repeat_one = value
        self.playback_flags_changed.fire()

    def get_queue_tracks(self):
        """
        Return :attr:`.Queue.get_tracks`
        """
        return self.queue.get_tracks()

    def _play(self):
        """
        Pick current track from a queue and requests media stream URL.
        Completes in background.
        """
        track = self.queue.get_current_track()
        if track is None:
            return
        self._is_loading = True
        track.get_url(callback=self._play_ready)
        self.broadcast_state()
        self.track_changed.fire(track)

    def _play_ready(self, url, error, track):
        """
        Called once track's media stream URL request completes.
        If *error* is ``None``, tell libVLC to play media by *url*.
        """
        self._is_loading = False
        if error:
            NotificationArea.notify('Failed to request media URL: {}'.format(str(error)))
            return
        assert track
        media = vlc.Media(url)
        self.media_player.set_media(media)

        self.media_player.play()

    @property
    def is_loading(self):
        """
        True if current libVLC state is :attr:`vlc.State.Playing`
        """
        return self._is_loading

    @property
    def is_playing(self):
        """
        True if current libVLC state is :attr:`vlc.State.Playing`
        """
        return self.media_player.get_state() == vlc.State.Playing

    def play_pause(self):
        """
        Toggle playback, i.e. play if paused or pause if playing.
        """
        if self.is_playing:
            self.media_player.pause()
        else:
            self.media_player.play()

    def get_play_progress(self):
        """
        Return current playback position in range ``[0;1]`` (``float``).
        """
        return self.media_player.get_position()

    def get_play_progress_seconds(self):
        """
        Return current playback position in seconds (``int``).
        """
        return int(self.media_player.get_position() * self.media_player.get_length() / 1000)

    def get_length_seconds(self):
        """
        Return currently played track's length in seconds (``int``).
        """
        return int(self.media_player.get_length() // 1000)

    def next(self, force=False):
        """
        Advance to next track in queue.
        See :meth:`.Queue.next`.
        """
        self.queue.next(force)
        self._play()

    def get_current_track(self):
        """
        Return currently played track.
        See :meth:`.Queue.get_current_track`.
        """
        return self.queue.get_current_track()

    def seek(self, delta):
        """
        Seek to relative position.
        *delta* must be a ``float`` in range ``[-1;1]``.
        """
        self.media_player.set_position(self.get_play_progress() + delta)

    def seek_absolute(self, position):
        """
        Seek to absolute position.
        *position* must be a ``float`` in range ``[0;1]``.
        """
        self.media_player.set_position(position)

    @staticmethod
    def get_equalizer_freqs():
        """
        Return a list of equalizer frequencies for each band.
        """
        return [
            vlc.libvlc_audio_equalizer_get_band_frequency(index)
            for index
            in range(vlc.libvlc_audio_equalizer_get_band_count())
        ]

    def get_equalizer_amps(self):
        """
        Return a list of equalizer amplifications for each band.
        """
        return [
            vlc.libvlc_audio_equalizer_get_amp_at_index(
                self.equalizer,
                index
            )
            for index
            in range(vlc.libvlc_audio_equalizer_get_band_count())
        ]

    def set_equalizer_value(self, index, amp):
        """
        Set equalizer amplification for specific band.
        """
        assert vlc.libvlc_audio_equalizer_set_amp_at_index(
            self.equalizer,
            amp,
            index
        ) == 0
        self.media_player.set_equalizer(self.equalizer)

    def set_equalizer_values(self, amps):
        """
        Set a list of equalizer amplifications for each band.
        """
        assert len(amps) == vlc.libvlc_audio_equalizer_get_band_count()
        for index, amp in enumerate(amps):
            assert vlc.libvlc_audio_equalizer_set_amp_at_index(
                self.equalizer,
                amp,
                index
            ) == 0
        self.media_player.set_equalizer(self.equalizer)



"""
Components for search page.
"""
import urwid

from clay.gp import GP
from clay.songlist import SongListBox
from clay.notifications import NotificationArea
from clay.pages.page import AbstractPage


class ArtistListBox(urwid.ListBox):
    """
    Widget that displays list of artists.
    """
    def __init__(self):
        self.walker = urwid.SimpleListWalker([])
        super(ArtistListBox, self).__init__(self.walker)


class SearchBox(urwid.Columns):
    """
    Widget that displays search input and results.
    """
    signals = ['search-requested']

    def __init__(self):
        self.query = urwid.Edit()
        super(SearchBox, self).__init__([
            ('pack', urwid.Text('Search: ')),
            urwid.AttrWrap(self.query, 'input', 'input_focus')
        ])

    def keypress(self, size, key):
        """
        Handle keypress.
        """
        if key == 'enter':
            urwid.emit_signal(self, 'search-requested', self.query.edit_text)
            return None
        return super(SearchBox, self).keypress(size, key)


class SearchPage(urwid.Pile, AbstractPage):
    """
    Search page.

    Allows to perform searches & displays search results.
    """
    @property
    def name(self):
        return 'Search'

    @property
    def key(self):
        return 4

    def __init__(self, app):
        self.app = app
        self.songlist = SongListBox(app)

        self.search_box = SearchBox()

        urwid.connect_signal(self.search_box, 'search-requested', self.perform_search)

        super(SearchPage, self).__init__([
            ('pack', self.search_box),
            ('pack', urwid.Divider(u'\u2500')),
            self.songlist
        ])

    def perform_search(self, query):
        """
        Search tracks by query.
        """
        self.songlist.set_placeholder(u' \U0001F50D Searching for "{}"...'.format(
            query
        ))
        GP.get().search_async(query, callback=self.search_finished)

    def search_finished(self, results, error):
        """
        Populate song list with search results.
        """
        if error:
            NotificationArea.notify('Failed to search: {}'.format(str(error)))
        else:
            self.songlist.populate(results.get_tracks())
            self.app.redraw()

    def activate(self):
        pass

    def keypress(self, size, key):
        if key == 'tab':
            if self.focus == self.search_box:
                self.focus_position = 2
            else:
                self.focus_position = 0
            return None
        else:
            return super(SearchPage, self).keypress(size, key)



"""
Library page.
"""
import urwid

from clay.gp import GP
from clay.songlist import SongListBox
from clay.notifications import NotificationArea
from clay.pages.page import AbstractPage


class MyLibraryPage(urwid.Columns, AbstractPage):
    """
    My library page.

    Displays :class:`clay.songlist.SongListBox` with all songs in library.
    """
    @property
    def name(self):
        return 'Library'

    @property
    def key(self):
        return 1

    def __init__(self, app):
        self.app = app
        self.songlist = SongListBox(app)
        self.notification = None

        GP.get().auth_state_changed += self.get_all_songs
        GP.get().caches_invalidated += self.get_all_songs

        super(MyLibraryPage, self).__init__([
            self.songlist
        ])

    def on_get_all_songs(self, tracks, error):
        """
        Called when all library songs are fetched from server.
        Populate song list.
        """
        if error:
            NotificationArea.notify('Failed to load my library: {}'.format(str(error)))
            return
        # self.notification.close()
        self.songlist.populate(tracks)
        self.app.redraw()

    def get_all_songs(self, *_):
        """
        Called when auth state changes or GP caches are invalidated.
        """
        if GP.get().is_authenticated:
            self.songlist.set_placeholder(u'\n \uf01e Loading song list...')

            GP.get().get_all_tracks_async(callback=self.on_get_all_songs)
            self.app.redraw()
        # self.notification = NotificationArea.notify('Loading library...')

    def activate(self):
        pass



#!/usr/bin/env python

from setuptools import setup, find_packages

setup(
    name='clay-player',
    version='0.5',
    description='Command Line Player for Google Play Music',
    author='Andrew Dunai',
    author_email='a@dun.ai',
    url='https://github.com/and3rson/clay',
    install_required=[
        'gmusicapi',
        'PyYAML',
        'urwid'
    ],
    packages=find_packages(),
    entry_points={
        'console_scripts': [
            'clay=clay.app:main'
        ]
    }
)




"""
Generic page classes.
"""


class AbstractPage(object):
    """
    Represents app page.
    """
    def name(self):
        """
        Return page name.
        """
        raise NotImplementedError()

    def key(self):
        """
        Return page key (``int``), used for hotkeys.
        """
        raise NotImplementedError()

    def activate(self):
        """
        Notify page that it is activated.
        """
        raise NotImplementedError()



"""
Notification widgets.
"""
import urwid


class Notification(urwid.Columns):
    """
    Single notification widget.
    Can be updated or closed.
    """
    TEMPLATE = u' \u26A1 {}'

    def __init__(self, area, notification_id, message):
        self.area = area
        self._id = notification_id
        self.text = urwid.Text('')
        self._set_text(message)
        super(Notification, self).__init__([
            urwid.AttrWrap(
                urwid.Columns([
                    self.text,
                    ('pack', urwid.Text('[Hit ESC to close] '))
                ]),
                'notification'
            )
        ])

    @property
    def id(self):  # pylint: disable=invalid-name
        """
        Notification ID.
        """
        return self._id

    def _set_text(self, message):
        """
        Set contents for this notification.
        """
        message = message.split('\n')
        message = '\n'.join([
            message[0]
        ] + ['    {}'.format(line) for line in message[1:]])
        self.text.set_text(Notification.TEMPLATE.format(message))

    def update(self, message):
        """
        Update notification message.
        """
        self._set_text(message)
        if not self.is_alive:
            self.area.append_notification(self)
        self.area.__class__.app.redraw()

    @property
    def is_alive(self):
        """
        Return True if notification is currently visible.
        """
        for notification, _ in self.area.contents:
            if notification is self:
                return True
        return False

    def close(self):
        """
        Close notification.
        """
        for notification, props in reversed(self.area.contents):
            if notification is self:
                self.area.contents.remove((notification, props))

        if self.area.__class__.app is not None:
            self.area.__class__.app.redraw()


class NotificationArea(urwid.Pile):
    """
    Notification area widget.
    """
    instance = None
    app = None

    def __init__(self):
        assert self.__class__.instance is None, 'Can be created only once!'
        self.last_id = 0
        self.notifications = {}
        super(NotificationArea, self).__init__([])

    @classmethod
    def get(cls):
        """
        Create new :class:`.NotificationArea` instance or return existing one.
        """
        if cls.instance is None:
            cls.instance = NotificationArea()

        return cls.instance

    @classmethod
    def set_app(cls, app):
        """
        Set app instance.

        Required for proper screen redraws when
        new notifications are created asynchronously.
        """
        cls.app = app

    @classmethod
    def notify(cls, message):
        """
        Create new notification with message.
        This is a class method.
        """
        return cls.get().do_notify(message)

    @classmethod
    def close_all(cls):
        """
        Close all notfiications.
        This is a class method.
        """
        cls.get().do_close_all()

    @classmethod
    def close_newest(cls):
        """
        Close newest notification.
        This is a class method.
        """
        cls.get().do_close_newest()

    def do_notify(self, message):
        """
        Create new notification with message.
        """
        self.last_id += 1
        notification = Notification(self, self.last_id, message)
        self.append_notification(notification)
        return notification

    def append_notification(self, notification):
        """
        Append an existing notification (that was probably closed).
        """
        self.contents.append(
            (
                notification,
                ('weight', 1)
            )
        )
        if self.__class__.app is not None:
            self.__class__.app.redraw()

    def do_close_all(self):
        """
        Close all notifications.
        """
        while self.contents:
            self.contents[0][0].close()

    def do_close_newest(self):
        """
        Close newest notification
        """
        if not self.contents:
            return
        self.contents[-1][0].close()



"""
Google Play Music integration via gmusicapi.
"""
# pylint: disable=broad-except
# pylint: disable=too-many-arguments
# pylint: disable=too-many-instance-attributes
from threading import Thread, Lock
from uuid import UUID

from gmusicapi.clients import Mobileclient

from clay.eventhook import EventHook


def asynchronous(func):
    """
    Decorates a function to become asynchronous.

    Once called, runs original function in a new Thread.

    Must be called with a 'callback' argument that will be called
    once thread with original function finishes. Receives two args:
    result and error.

    - "result" contains function return value or None if there was an exception.
    - "error" contains None or Exception if there was one.
    """
    def wrapper(*args, **kwargs):
        """
        Inner function.
        """
        callback = kwargs.pop('callback')
        extra = kwargs.pop('extra', dict())

        def process():
            """
            Thread body.
            """
            try:
                result = func(*args, **kwargs)
            except Exception as error:
                callback(None, error, **extra)
            else:
                callback(result, None, **extra)

        Thread(target=process).start()
    return wrapper


def synchronized(func):
    """
    Decorates a function to become thread-safe by preventing
    it from being executed multiple times before previous calls end.

    Lock is acquired on entrance and is released on return or Exception.
    """
    lock = Lock()

    def wrapper(*args, **kwargs):
        """
        Inner function.
        """
        try:
            lock.acquire()
            return func(*args, **kwargs)
        finally:
            lock.release()

    return wrapper


class Track(object):
    """
    Model that represents single track from Google Play Music.
    """
    TYPE_UPLOADED = 'uploaded'
    TYPE_STORE = 'store'

    SOURCE_LIBRARY = 'library'
    SOURCE_STATION = 'station'
    SOURCE_PLAYLIST = 'playlist'
    SOURCE_SEARCH = 'search'

    def __init__(
            self,
            title, artist, duration, source,
            library_id=None, store_id=None, playlist_item_id=None,
            album_name=None, album_url=None
    ):
        self.title = title
        self.artist = artist
        self.duration = duration
        self.source = source

        self.cached_url = None

        self.library_id = library_id
        self.store_id = store_id
        self.playlist_item_id = playlist_item_id

        self.album_name = album_name
        self.album_url = album_url

    @property
    def id(self):  # pylint: disable=invalid-name
        """
        Return ID for this track.
        """
        if self.library_id:
            return self.library_id
        return self.store_id

    def __eq__(self, other):
        return (
            (self.library_id and self.library_id == other.library_id) or
            (self.store_id and self.store_id == other.store_id) or
            (self.playlist_item_id and self.playlist_item_id == other.playlist_item_id)
        )

    @classmethod
    def from_data(cls, data, source, many=False):
        """
        Construct and return one or many :class:`.Track` instances
        from Google Play Music API response.
        """
        if many:
            return [cls.from_data(one, source) for one in data]

        if source == Track.SOURCE_SEARCH:
            # Data contains a nested track representation.
            return Track(
                title=data['track']['title'],
                artist=data['track']['artist'],
                duration=int(data['track']['durationMillis']),
                source=source,
                store_id=data['track']['storeId'],  # or data['trackId']
                album_name=data['track']['album'],
                album_url=data['track']['albumArtRef'][0]['url']
            )
        elif source == Track.SOURCE_STATION:
            # Station tracks have all the info in place.
            return Track(
                title=data['title'],
                artist=data['artist'],
                duration=int(data['durationMillis']),
                source=source,
                store_id=data['storeId'],
                album_name=data['album'],
                album_url=data['albumArtRef'][0]['url']
            )
        elif source == Track.SOURCE_LIBRARY:
            # Data contains all info about track
            # including ID in library and ID in store.
            UUID(data['id'])
            return Track(
                title=data['title'],
                artist=data['artist'],
                duration=int(data['durationMillis']),
                source=source,
                store_id=data['storeId'],
                library_id=data['id'],
                album_name=data['album'],
                album_url=data['albumArtRef'][0]['url']
            )
        elif source == Track.SOURCE_PLAYLIST:
            if 'track' in data:
                # Data contains a nested track representation that can be used
                # to construct new track.
                return Track(
                    title=data['track']['title'],
                    artist=data['track']['artist'],
                    duration=int(data['track']['durationMillis']),
                    source=source,
                    store_id=data['track']['storeId'],  # or data['trackId']
                    playlist_item_id=data['id'],
                    album_name=data['track']['album'],
                    album_url=data['track']['albumArtRef'][0]['url']
                )
            # We need to find a track in Library by trackId.
            UUID(data['trackId'])
            track = GP.get().get_track_by_id(data['trackId'])
            return Track(
                title=track.title,
                artist=track.artist,
                duration=track.duration,
                source=source,
                store_id=track.store_id,
                album_name=track.album_name,
                album_url=track.album_url
            )
        raise AssertionError()

    def get_url(self, callback):
        """
        Gets playable stream URL for this track.

        "callback" is called with "(url, error)" args after URL is fetched.

        Keep in mind this URL is valid for a limited time.
        """
        def on_get_url(url, error):
            """
            Called when URL is fetched.
            """
            self.cached_url = url
            callback(url, error, self)

        GP.get().get_stream_url_async(self.store_id, callback=on_get_url)

    @synchronized
    def create_station(self):
        """
        Creates a new station from this :class:`.Track`.

        Returns :class:`.Station` instance.
        """
        station_id = GP.get().mobile_client.create_station(
            name=u'Station - {}'.format(self.title),
            track_id=self.store_id
        )
        station = Station(station_id)
        station.load_tracks()
        return station

    create_station_async = asynchronous(create_station)

    def add_to_my_library(self):
        """
        Add a track to my library.
        """
        return GP.get().add_to_my_library(self)

    add_to_my_library_async = asynchronous(add_to_my_library)

    def remove_from_my_library(self):
        """
        Remove a track from my library.
        """
        return GP.get().remove_from_my_library(self)

    remove_from_my_library_async = asynchronous(remove_from_my_library)

    def __str__(self):
        return u'<Track "{} - {}" from {}>'.format(
            self.artist,
            self.title,
            self.source
        )

    __repr__ = __str__


class Artist(object):
    """
    Model that represents artist.
    """
    def __init__(self, artist_id, name):
        self._id = artist_id
        self.name = name

    @property
    def id(self):  # pylint: disable=invalid-name
        """
        Artist ID.
        """
        return self._id

    @classmethod
    def from_data(cls, data, many=False):
        """
        Construct and return one or many :class:`.Artist` instances
        from Google Play Music API response.
        """
        if many:
            return [cls.from_data(one) for one in data]

        return Artist(
            artist_id=data['artistId'],
            name=data['name']
        )


class Station(object):
    """
    Model that represents specific station on Google Play Music.
    """
    def __init__(self, station_id):
        self._id = station_id
        self._tracks = []
        self._tracks_loaded = False

    @property
    def id(self):  # pylint: disable=invalid-name
        """
        Station ID.
        """
        return self._id

    def load_tracks(self):
        """
        Fetch tracks related to this station and
        populate it with :class:`Track` instances.
        """
        data = GP.get().mobile_client.get_station_tracks(self.id, 100)
        self._tracks = Track.from_data(data, Track.SOURCE_STATION, many=True)
        self._tracks_loaded = True

    def get_tracks(self):
        """
        Return a list of tracks in this station.
        """
        assert self._tracks_loaded, 'Must call ".load_tracks()" before ".get_tracks()"'
        return self._tracks


class SearchResults(object):
    """
    Model that represents search results including artists & tracks.
    """
    def __init__(self, tracks, artists):
        self.artists = artists
        self.tracks = tracks

    @classmethod
    def from_data(cls, data):
        """
        Construct and return :class:`.SearchResults` instance from raw data.
        """
        return SearchResults(
            tracks=Track.from_data(data['song_hits'], Track.SOURCE_SEARCH, many=True),
            artists=Artist.from_data([
                item['artist']
                for item
                in data['artist_hits']
            ], many=True)
        )

    def get_artists(self):
        """
        Return found artists.
        """
        return self.artists

    def get_tracks(self):
        """
        Return found tracks.
        """
        return self.tracks


class Playlist(object):
    """
    Model that represents remotely stored (Google Play Music) playlist.
    """
    def __init__(self, playlist_id, name, tracks):
        self._id = playlist_id
        self.name = name
        self.tracks = tracks

    @property
    def id(self):  # pylint: disable=invalid-name
        """
        Playlist ID.
        """
        return self._id

    @classmethod
    def from_data(cls, data, many=False):
        """
        Construct and return one or many :class:`.Playlist` instances
        from Google Play Music API response.
        """
        if many:
            return [cls.from_data(one) for one in data]

        return Playlist(
            playlist_id=data['id'],
            name=data['name'],
            tracks=Track.from_data(data['tracks'], Track.SOURCE_PLAYLIST, many=True)
        )


class GP(object):
    """
    Interface to :class:`gmusicapi.Mobileclient`. Implements
    asynchronous API calls, caching and some other perks.

    Singleton.
    """
    # TODO: Switch to urwid signals for more explicitness?
    instance = None

    caches_invalidated = EventHook()

    def __init__(self):
        assert self.__class__.instance is None, 'Can be created only once!'
        self.mobile_client = Mobileclient()
        self.cached_tracks = None
        self.cached_playlists = None

        self.invalidate_caches()

        self.auth_state_changed = EventHook()

    @classmethod
    def get(cls):
        """
        Create new :class:`.GP` instance or return existing one.
        """
        if cls.instance is None:
            cls.instance = GP()

        return cls.instance

    def invalidate_caches(self):
        """
        Clear cached tracks & playlists.
        """
        self.cached_tracks = None
        self.cached_playlists = None
        self.caches_invalidated.fire()

    @synchronized
    def login(self, email, password, device_id, **_):
        """
        Log in into Google Play Music.
        """
        self.mobile_client.logout()
        self.invalidate_caches()
        # prev_auth_state = self.is_authenticated
        result = self.mobile_client.login(email, password, device_id)
        # if prev_auth_state != self.is_authenticated:
        self.auth_state_changed.fire(self.is_authenticated)
        return result

    login_async = asynchronous(login)

    @synchronized
    def use_authtoken(self, authtoken, device_id):
        """
        Try to use cached token to log into Google Play Music.
        """
        # pylint: disable=protected-access
        self.mobile_client.session._authtoken = authtoken
        self.mobile_client.session.is_authenticated = True
        self.mobile_client.android_id = device_id
        del self.mobile_client.is_subscribed
        if self.mobile_client.is_subscribed:
            self.auth_state_changed.fire(True)
            return True
        del self.mobile_client.is_subscribed
        self.mobile_client.android_id = None
        self.mobile_client.session.is_authenticated = False
        self.auth_state_changed.fire(False)
        return False

    use_authtoken_async = asynchronous(use_authtoken)

    def get_authtoken(self):
        """
        Return currently active auth token.
        """
        # pylint: disable=protected-access
        return self.mobile_client.session._authtoken

    @synchronized
    def get_all_tracks(self):
        """
        Cache and return all tracks from "My library".

        Each track will have "id" and "storeId" keys.
        """
        if self.cached_tracks:
            return self.cached_tracks
        data = self.mobile_client.get_all_songs()
        self.cached_tracks = Track.from_data(data, Track.SOURCE_LIBRARY, True)
        return self.cached_tracks

    get_all_tracks_async = asynchronous(get_all_tracks)

    def get_stream_url(self, stream_id):
        """
        Returns playable stream URL of track by id.
        """
        return self.mobile_client.get_stream_url(stream_id)

    get_stream_url_async = asynchronous(get_stream_url)

    @synchronized
    def get_all_user_playlist_contents(self, **_):
        """
        Return list of :class:`.Playlist` instances.
        """
        if self.cached_playlists:
            return self.cached_playlists
        self.get_all_tracks()

        self.cached_playlists = Playlist.from_data(
            self.mobile_client.get_all_user_playlist_contents(),
            True
        )
        return self.cached_playlists

    get_all_user_playlist_contents_async = (  # pylint: disable=invalid-name
        asynchronous(get_all_user_playlist_contents)
    )

    def get_cached_tracks_map(self):
        """
        Return a dictionary of tracks where keys are strings with track IDs
        and values are :class:`.Track` instances.
        """
        return {track.id: track for track in self.cached_tracks}

    def get_track_by_id(self, any_id):
        """
        Return track by id or store_id.
        """
        for track in self.cached_tracks:
            if any_id in (track.library_id, track.store_id, track.playlist_item_id):
                return track
        return None

    def search(self, query):
        """
        Find tracks and return an instance of :class:`.SearchResults`.
        """
        results = self.mobile_client.search(query)
        return SearchResults.from_data(results)

    search_async = asynchronous(search)

    def add_to_my_library(self, track):
        """
        Add a track to my library.
        """
        result = self.mobile_client.add_store_tracks(track.id)
        if result:
            self.invalidate_caches()
        return result

    def remove_from_my_library(self, track):
        """
        Remove a track from my library.
        """
        result = self.mobile_client.delete_songs(track.id)
        if result:
            self.invalidate_caches()
        return result

    @property
    def is_authenticated(self):
        """
        Return True if user is authenticated on Google Play Music, false otherwise.
        """
        return self.mobile_client.is_authenticated()
